@@ -5,7 +5,7 @@
 // http://opensource.org/licenses/MIT>, at your option. This file may not be
 // copied, modified, or distributed except according to those terms.
 
-///! Macros for static_init crate.
+/// Macros for static_init crate.
 extern crate proc_macro;
 extern crate syn;
 use syn::spanned::Spanned;
@@ -14,7 +14,7 @@ use syn::*;
 use core::result::Result;
 
 extern crate quote;
-use quote::quote_spanned;
+use quote::{quote, quote_spanned};
 
 use proc_macro::TokenStream;
 
@@ -34,109 +34,217 @@ macro_rules! ok_or_return {
 pub fn constructor(args: TokenStream, input: TokenStream) -> TokenStream {
     let priority = ok_or_return!(parse_priority(args));
 
-    let section = ok_or_return!(init_section(priority));
+    let sections = init_section_map(priority);
 
     let func: ItemFn = parse_macro_input!(input);
 
-    let func_ptr_name = format!("__static_init_constructor_{}", func.sig.ident);
+    let func_ptr_name = format!("__STATIC_INIT_CONSTRUCTOR_{}", func.sig.ident);
 
     let func_type = get_init_func_sig(&func.sig);
 
-    gen_ctor_dtor(func, &section, &func_ptr_name, func_type).into()
+    gen_ctor_dtor(func, sections, &func_ptr_name, func_type).into()
 }
 
 fn get_init_func_sig(sig: &Signature) -> TypeBareFn {
     let sp = sig.span().resolved_at(Span::mixed_site());
 
     if cfg!(target_env = "gnu") && cfg!(target_family = "unix") && !sig.inputs.is_empty() {
-        parse2(quote_spanned!(sp=>extern "C" fn(i32,*const*const u8, *const *const u8)))
-            .unwrap()
+        parse2(quote_spanned!(sp=>extern "C" fn(i32,*const*const u8, *const *const u8))).unwrap()
     } else {
         parse2(quote_spanned!(sp=>extern "C" fn())).unwrap()
     }
 }
 
-fn const_dtor_no_support() -> TokenStream {
+fn const_dtor_no_support() -> TokenStream2 {
     quote_spanned!(Span::mixed_site()=>compile_error!(
         "program constructors/destructors not supported on this target"
     ))
-    .into()
 }
 
-fn init_section(priority: u16) -> Result<String, TokenStream> {
+fn gen_cfg_elf() -> TokenStream2 {
+    quote! {
+        any(
+            target_os = "linux",
+            target_os = "android",
+            target_os = "freebsd",
+            target_os = "dragonfly",
+            target_os = "netbsd",
+            target_os = "openbsd",
+            target_os = "solaris",
+            target_os = "illumos",
+            target_os = "emscripten",
+            target_os = "haiku",
+            target_os = "l4re",
+            target_os = "fuchsia",
+            target_os = "redox",
+            target_os = "vxworks"
+        )
+    }
+}
+
+fn gen_cfg_mach_o() -> TokenStream2 {
+    quote! {
+        any(target_os = "macos", target_os = "ios")
+    }
+}
+
+fn gen_cfg_coff() -> TokenStream2 {
+    quote! {
+        target_os = "windows"
+    }
+}
+
+fn gen_cfg_constructor_destructor() -> TokenStream2 {
+    let elf = gen_cfg_elf();
+    let coff = gen_cfg_coff();
+    let mach_o = gen_cfg_mach_o();
+    quote! {
+        any(#elf, #coff, #mach_o)
+    }
+}
+
+fn init_section_map(priority: u16) -> Vec<(TokenStream2, Result<String, TokenStream2>)> {
     //Todo priority bellow(65535-100) should be unsafe
     //or increase the number to be above 100
     //
     //idem put Room for lesser lazy initialized later
-    if cfg!(elf) {
+
+    let elf = gen_cfg_elf();
+    let mach_o = gen_cfg_mach_o();
+    let coff = gen_cfg_coff();
+    let mut mapping = vec![];
+
+    mapping.push((
+        quote! {
+            not(any(#elf, #mach_o, #coff))
+        },
+        Err(const_dtor_no_support())
+    ));
+
+    mapping.push((
+        elf,
         // on linux the standard library args are initilized
         // at .init_array.00099. => priority 65436
         Ok(format!(".init_array.{:05}", 65535 - priority))
-    } else if cfg!(mach_o) {
+    ));
+
+    mapping.push((
+        mach_o,
         //on mach it is not clear of ObjC runtime is initialized
         //before or after constructors that are here
         if priority != 0 {
             Err(quote_spanned!(Span::mixed_site()=>compile_error!(
-                "Constructor priority other than 0 not supported on this plateform."
-            ))
-            .into())
+                "Constructor priority other than 0 not supported on this platform."
+            )))
         } else {
             Ok("__DATA,__mod_init_func".to_string())
         }
-    } else if cfg!(coff) {
+    ));
+
+    mapping.push((
+        coff,
         // on windows init maybe be called at .CRT$XCU
         // so lets initialization takes place after
         Ok(format!(".CRT$XCU{:05}", 65535 - priority))
-    } else {
-        Err(const_dtor_no_support())
-    }
+    ));
+
+    mapping
 }
 
-fn fini_section(priority: u16) -> Result<String, TokenStream> {
-    if cfg!(elf) {
+
+
+fn fini_section_map(priority: u16) -> Vec<(TokenStream2, Result<String, TokenStream2>)> {
+    let elf = gen_cfg_elf();
+    let mach_o = gen_cfg_mach_o();
+    let coff = gen_cfg_coff();
+    let mut mapping = vec![];
+
+    mapping.push((
+        quote! {
+            not(any(#elf, #mach_o, #coff))
+        },
+        Err(const_dtor_no_support())
+    ));
+
+    mapping.push((
+        elf,
         // destructors not used by standard library
         Ok(format!(".fini_array.{:05}", 65535 - priority))
-    } else if cfg!(mach_o) {
+    ));
+
+    mapping.push((
+        mach_o,
         if priority != 0 {
             Err(quote_spanned!(Span::mixed_site()=>compile_error!(
-                "Constructor priority not supported on this plateform."
-            ))
-            .into())
+                "Constructor priority not supported on this platform."
+            )))
         } else {
             Ok("__DATA,__mod_term_func".to_string())
         }
-    } else if cfg!(coff) {
+    ));
+
+    mapping.push((
+        coff,
         // destructors not used by standard library
         Ok(format!(".CRT$XPTZ{:05}", 65535 - priority))
-    } else {
-        Err(const_dtor_no_support())
-    }
+    ));
+
+    mapping
 }
 
+
 #[proc_macro_attribute]
 pub fn destructor(args: TokenStream, input: TokenStream) -> TokenStream {
     let priority = ok_or_return!(parse_priority(args));
 
-    let section = ok_or_return!(fini_section(priority));
+    let sections = fini_section_map(priority);
 
     let func: ItemFn = parse_macro_input!(input);
 
-    let func_ptr_name = format!("__static_init_destructor_{}", func.sig.ident);
+    let func_ptr_name = format!("__STATIC_INIT_DESTRUCTOR_{}", func.sig.ident);
 
     let sp = func.sig.span();
     let func_type = parse2(quote_spanned!(sp.span()=>extern "C" fn())).unwrap();
 
-    gen_ctor_dtor(func, &section, &func_ptr_name, func_type).into()
+    gen_ctor_dtor(func, sections, &func_ptr_name, func_type).into()
 }
 
-
 #[proc_macro_attribute]
 pub fn dynamic(args: TokenStream, input: TokenStream) -> TokenStream {
     let item: ItemStatic = parse_macro_input!(input);
 
-    let options = ok_or_return!(parse_dyn_options(parse_macro_input!(args)));
+    let constructor_destructor = gen_cfg_constructor_destructor();
 
-    gen_dyn_init(item, options).into()
+    let args: Vec<NestedMeta> = parse_macro_input!(args);
+
+    let cfg_constructor_destructor = quote! {
+        #[cfg(#constructor_destructor)]
+    };
+
+    let tt_constructor_destructor = match parse_dyn_options(args.clone(), true) {
+        Ok(options) => gen_dyn_init(&item, options, &cfg_constructor_destructor),
+        Err(err) => quote! {
+            #cfg_constructor_destructor
+            #err;
+        }
+    };
+
+    let cfg_no_constructor_destructor = quote! {
+        #[cfg(not(#constructor_destructor))]
+    };
+
+    let tt_no_constructor_destructor = match parse_dyn_options(args, false) {
+        Ok(options) => gen_dyn_init(&item, options, &cfg_no_constructor_destructor),
+        Err(err) => quote! {
+            #cfg_no_constructor_destructor
+            #err;
+        }
+    };
+
+    quote::quote! {
+        #tt_constructor_destructor
+        #tt_no_constructor_destructor
+    }.into()
 }
 
 #[derive(Clone, Copy, Eq, PartialEq)]
@@ -155,16 +263,16 @@ enum DropMode {
 }
 #[derive(Clone, Copy, Eq, PartialEq)]
 struct Tolerance {
-    init_fail:         bool,
+    init_fail: bool,
     registration_fail: bool,
 }
 
 #[derive(Clone, Copy, Eq, PartialEq)]
 struct DynMode {
-    init:      InitMode,
-    drop:      DropMode,
+    init: InitMode,
+    drop: DropMode,
     tolerance: Tolerance,
-    priming:   bool,
+    priming: bool,
 }
 
 fn parse_priority(args: TokenStream) -> std::result::Result<u16, TokenStream2> {
@@ -216,15 +324,15 @@ fn parse_priority_literal(lit: &Lit) -> Result<u16, TokenStream2> {
     }
 }
 
-fn parse_dyn_options(args: AttributeArgs) -> std::result::Result<DynMode, TokenStream2> {
+fn parse_dyn_options(args: AttributeArgs, constructor_destructor: bool) -> std::result::Result<DynMode, TokenStream2> {
     let mut opt = DynMode {
-        init:      InitMode::LesserLazy,
-        drop:      DropMode::None,
+        init: InitMode::LesserLazy,
+        drop: DropMode::None,
         tolerance: Tolerance {
-            init_fail:         true,
+            init_fail: true,
             registration_fail: false,
         },
-        priming:   false,
+        priming: false,
     };
 
     let mut init_set = false;
@@ -274,14 +382,18 @@ fn parse_dyn_options(args: AttributeArgs) -> std::result::Result<DynMode, TokenS
                     check_no_init!(id);
                     opt.init = InitMode::Dynamic(0);
                 } else if id == "drop" {
-                    if !cfg!(constructor_destructor) {
-                        return Err(generate_error!(id.span()=>"Static drop mode only supported on unixes and windows"))
+                    if !constructor_destructor {
+                        return Err(
+                            generate_error!(id.span()=>"Static drop mode only supported on unixes and windows"),
+                        );
                     }
                     check_no_drop!(id);
                     opt.drop = DropMode::Drop;
                 } else if id == "finalize" {
-                    if !cfg!(constructor_destructor) {
-                        return Err(generate_error!(id.span()=>"Static finalization mode only supported on unixes and windows"))
+                    if !constructor_destructor {
+                        return Err(
+                            generate_error!(id.span()=>"Static finalization mode only supported on unixes and windows"),
+                        );
                     }
                     check_no_drop!(id);
                     opt.drop = DropMode::Finalize;
@@ -335,7 +447,7 @@ fn parse_dyn_options(args: AttributeArgs) -> std::result::Result<DynMode, TokenS
             }
         }
     }
-    if opt.init == InitMode::LesserLazy && !cfg!(constructor_destructor) {
+    if opt.init == InitMode::LesserLazy && !constructor_destructor {
         opt.init = InitMode::Lazy
     }
     if opt.drop == DropMode::None && opt.tolerance.registration_fail {
@@ -344,10 +456,8 @@ fn parse_dyn_options(args: AttributeArgs) -> std::result::Result<DynMode, TokenS
              `drop` or `finalize` attribute argument if the intent is that this static is dropped."
         ));
     }
-    if opt.priming && ! (opt.init== InitMode::Lazy || opt.init == InitMode::LesserLazy) {
-        return Err(generate_error!(
-            "Only lazy statics can be primed"
-        ));
+    if opt.priming && !(opt.init == InitMode::Lazy || opt.init == InitMode::LesserLazy) {
+        return Err(generate_error!("Only lazy statics can be primed"));
     }
     if (opt.init == InitMode::Lazy || opt.init == InitMode::LesserLazy)
         && !(opt.drop == DropMode::None
@@ -389,29 +499,44 @@ fn parse_dyn_options(args: AttributeArgs) -> std::result::Result<DynMode, TokenS
 
 fn gen_ctor_dtor(
     func: ItemFn,
-    section: &str,
+    sections: Vec<(TokenStream2, Result<String, TokenStream2>)>,
     func_ptr_name: &str,
     typ: TypeBareFn,
 ) -> TokenStream2 {
     let func_ptr_name = Ident::new(func_ptr_name, Span::call_site());
 
-    let section = LitStr::new(section, Span::call_site());
-
     let func_name = &func.sig.ident;
 
     let sp = func.sig.span().resolved_at(Span::mixed_site());
-    //if func.sig.unsafety.is_none() {
-    //    quote_spanned! {sp=>compile_error!("Constructors and destructors must be unsafe functions as \
-    //    they may access uninitialized memory regions")}
-    //} else {
-    quote_spanned! {sp=>
+
+    let statics = sections.into_iter().map(|(cfg, section)| {
+        match section {
+            Ok(section) => {
+                let section = LitStr::new(&section, Span::call_site());
+
+                quote_spanned! {sp=>
+                    #[cfg(#cfg)]
+                    #[doc(hidden)]
+                    #[link_section = #section]
+                    #[used]
+                    #[allow(non_upper_case_globals)]
+                    pub static #func_ptr_name: #typ = #func_name;
+                }
+            }
+            Err(err) => {
+                quote_spanned!(sp=>
+                    #[cfg(#cfg)]
+                    #err;
+                )
+            }
+        }
+    }).collect::<Vec<_>>();
+
+    quote_spanned!(sp=>
         #func
-        #[doc(hidden)]
-        #[link_section = #section]
-        #[used]
-        pub static #func_ptr_name: #typ = #func_name;
-    }
-    //}
+
+        #(#statics)*
+    )
 }
 
 fn has_thread_local(attrs: &[Attribute]) -> bool {
@@ -425,11 +550,12 @@ fn has_thread_local(attrs: &[Attribute]) -> bool {
     false
 }
 
-fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
+fn gen_dyn_init(stat: &ItemStatic, options: DynMode, cfg: &TokenStream2) -> TokenStream2 {
     //TODO: dropped static must be initialized by unsafe code because
     //if initialization panic this will cause UB TBC.
     //
     //TODO: for lazy statics with leak tolerance => only usefull for thread locals.
+    let mut stat = stat.clone();
 
     let stat_name = &stat.ident;
     let stat_vis = &stat.vis;
@@ -442,7 +568,6 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
         "Expected an expression of the form `match INIT { PRIME => /*expr/*, DYN => /*expr*/}`"
     );
 
-
     let stat_typ = &*stat.ty;
 
     let is_thread_local = has_thread_local(&stat.attrs);
@@ -523,7 +648,10 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
                 ::static_init::lazy::PrimedLockedLazy::<#stat_typ,#stat_generator_name>
             }
         }
-    } else if options.priming && options.init == InitMode::LesserLazy && options.drop == DropMode::None {
+    } else if options.priming
+        && options.init == InitMode::LesserLazy
+        && options.drop == DropMode::None
+    {
         if stat.mutability.is_none() {
             return generate_error!(stat.static_token.span()=>
                 "Primed statics are mutating (safe). Add the `mut` keyword."
@@ -534,7 +662,7 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
                 ::static_init::lazy::PrimedLesserLockedLazy::<#stat_typ,#stat_generator_name>
             }
         }
-    } else if options.priming  && options.init == InitMode::Lazy{
+    } else if options.priming && options.init == InitMode::Lazy {
         if stat.mutability.is_none() {
             return generate_error!(stat.static_token.span()=>
                 "Primed statics are mutating (safe). Add the `mut` keyword."
@@ -545,7 +673,7 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
                 ::static_init::lazy::PrimedLockedLazyDroped::<#stat_typ,#stat_generator_name>
             }
         }
-    } else if options.priming  && options.init == InitMode::LesserLazy{
+    } else if options.priming && options.init == InitMode::LesserLazy {
         if stat.mutability.is_none() {
             return generate_error!(stat.static_token.span()=>
                 "Primed statics are mutating (safe). Add the `mut` keyword."
@@ -707,39 +835,6 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
     let sp = stat.expr.span().resolved_at(Span::mixed_site());
 
     let initer = match options.init {
-        //InitMode::Dynamic(priority) if options.drop == DropMode::Drop => {
-        //    let attr: Attribute = parse_quote_spanned!(Span::mixed_site()=>
-        //    #[::static_init::constructor(#priority)]);
-        //    Some(quote_spanned! {sp=>
-        //            extern "C" fn __static_init_dropper() {
-        //                unsafe{#typ::drop(#stat_ref)}
-        //            }
-        //            #attr
-        //            extern "C" fn __static_init_initializer() {
-        //                ::static_init::raw_static::__set_init_prio(#priority as i32);
-        //                let __static_init_expr_result = #expr;
-        //                unsafe {#typ::set_to(#stat_ref,__static_init_expr_result);
-        //                ::libc::atexit(__static_init_dropper)};
-        //                ::static_init::raw_static::__set_init_prio(i32::MIN);
-        //            }
-        //    })
-        //}
-        //InitMode::Dynamic(priority) if options.drop == DropMode::Finalize => {
-        //    let attr: Attribute = parse_quote_spanned!(Span::mixed_site()=>#[::static_init::constructor(#priority)]);
-        //    Some(quote_spanned! {sp=>
-        //            extern "C" fn __static_init_dropper() {
-        //                unsafe{::static_init::Finaly::finalize(**#stat_ref)}
-        //            }
-        //            #attr
-        //            extern "C" fn __static_init_initializer() {
-        //                ::static_init::raw_static::__set_init_prio(#priority as i32);
-        //                let __static_init_expr_result = #expr;
-        //                unsafe {#typ::set_to(#stat_ref,__static_init_expr_result);
-        //                ::libc::atexit(__static_init_dropper)};
-        //                ::static_init::raw_static::__set_init_prio(i32::MIN);
-        //            }
-        //    })
-        //}
         InitMode::Dynamic(priority) => {
             let attr: Attribute = parse_quote!(#[::static_init::constructor(#priority)]);
             Some(quote_spanned! {sp=>
@@ -753,8 +848,16 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
             })
         }
 
-        InitMode::LesserLazy if !is_thread_local && cfg!(support_priority) => {
+        InitMode::LesserLazy if !is_thread_local => {
+            let support_priority = {
+                let elf = gen_cfg_elf();
+                let coff = gen_cfg_coff();
+                quote! {
+                    any(#elf, #coff)
+                }
+            };
             Some(quote_spanned! {sp=>
+                    #[cfg(#support_priority)]
                     #[::static_init::constructor(__lazy_init)]
                     extern "C" fn __static_init_initializer() {
                         unsafe {#typ::init(#stat_ref)};
@@ -812,14 +915,18 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
 
     let lazy_generator = if matches!(options.init, InitMode::Lazy | InitMode::LesserLazy) {
         Some(quote_spanned! {sp=>
+            #cfg
             #[allow(clippy::upper_case_acronyms)]
+            #[allow(non_camel_case_types)]
             #stat_vis struct #stat_generator_name;
+            #cfg
             impl ::static_init::Generator<#stat_typ> for #stat_generator_name {
                 #[inline]
                 fn generate(&self) -> #stat_typ {
                     #expr
                 }
             }
+            #cfg
             impl ::static_init::GeneratorTolerance for #stat_generator_name {
                 const INIT_FAILURE: bool = #init_fail_tol;
                 const FINAL_REGISTRATION_FAILURE: bool = #reg_fail_tol;
@@ -897,6 +1004,7 @@ fn gen_dyn_init(mut stat: ItemStatic, options: DynMode) -> TokenStream2 {
 
     quote_spanned! {sp=>
     #lazy_generator
+    #cfg
     #stat
     }
 }
